"""A shared collection of tools for general use."""

def ihelp(function_or_mod, show_help=True, show_code=True,return_code=False,markdown=True,file_location=False):
    """Call on any module or functon to display the object's
    help command printout AND/OR soruce code displayed as Markdown
    using Python-syntax"""
    import inspect
    
    try:
        from IPython import display, Markdown
    except:
        print('[!] IPython was not found.')
        
    page_header = '---'*28
    footer = '---'*28+'\n'
    if show_help:
        print(page_header)
        banner = ''.join(["---"*2,' HELP ',"---"*24,'\n'])
        print(banner)
        help(function_or_mod)
        # print(footer)
        
    import sys
    if "google.colab" in sys.modules:
        markdown=False

    if show_code:
        print(page_header)

        banner = ''.join(["---"*2,' SOURCE -',"---"*23])
        print(banner)
        try:
            import inspect
            source_DF = inspect.getsource(function_or_mod)

            if markdown == True:
                
                output = "```python" +'\n'+source_DF+'\n'+"```"
                display(Markdown(output))
            else:
                print(source_DF)

        except TypeError:
            pass
            # display(Markdown)


    if file_location:
        file_loc = inspect.getfile(function_or_mod)
        banner = ''.join(["---"*2,' FILE LOCATION ',"---"*21])
        print(page_header)
        print(banner)
        print(file_loc)

    # print(footer)

    if return_code:
        return source_DF


def list2df(list, index_col=None, set_caption=None, return_df=True,df_kwds=None): #, sort_values='index'):
    
    """ Quick turn an appened list with a header (row[0]) into a pretty dataframe.

        
        Args
            list (list of lists):
            index_col (string): name of column to set as index; None (Default) has integer index.
            set_caption (string):
            show_and_return (bool):
    
    EXAMPLE USE:
    >> list_results = [["Test","N","p-val"]] 
    
    # ... run test and append list of result values ...
    
    >> list_results.append([test_Name,length(data),p])
    
    ## Displays styled dataframe if caption:
    >> df = list2df(list_results, index_col="Test",
                     set_caption="Stat Test for Significance")
    

    """
    from IPython.display import display
    import pandas as pd
    df_list = pd.DataFrame(list[1:],columns=list[0],**df_kwds)
    
        
    if index_col is not None:
        df_list.reset_index(inplace=True)
        df_list.set_index(index_col, inplace=True)
        
    if set_caption is not None:
        dfs = df_list.style.set_caption()
        display(dfs)
    return df_list

# def ihelp_menu(function_list, json_file='ihelp_output.txt',to_embed=False):
#     """Accepts a list of string names for loaded modules/functions to save the `help` output and 
#     inspect.getsource() outputs to dictionary for later reference and display"""
#     ## One way using sys to write txt file
#     import pandas as pd
#     import sys
#     import inspect
#     from io import StringIO
#     notebook_output = sys.stdout
#     result = StringIO()
#     sys.stdout=result
    
#     ## Turn single input into a list
#     if isinstance(function_list,list)==False:
#         function_list = [function_list]
    
#     ## Make a dictionary of{function_name : function_object}
#     functions_dict = dict()
#     for fun in function_list:
        
#         ## if input is a string, save string as key, and eval(function) as value
#         if isinstance(fun, str):
#             functions_dict[fun] = eval(fun)

#         ## if input is a function, get the name of function using inspect and make key, function as value
#         elif inspect.isfunction(fun):

#             members= inspect.getmembers(fun)
#             member_df = pd.DataFrame(members,columns=['param','values']).set_index('param')

#             fun_name = member_df.loc['__name__'].values[0]
#             functions_dict[fun_name] = fun
            
            
#     ## Create an output dict to store results for functions
#     output_dict = {}

#     for fun_name, real_func in functions_dict.items():
        
#         output_dict[fun_name] = {}
                
#         ## First save help
#         help(real_func)
#         output_dict[fun_name]['help'] = result.getvalue()
        
#         ## Clear contents of io stream
#         result.truncate(0)
        
#         try:
#             ## Next save source
#             print(inspect.getsource(real_func)) #eval(fun)))###f"{eval(fun)}"))
#         except:
#             print("Source code for object was not found")
#         output_dict[fun_name]['source'] = result.getvalue()
        
#         ## clear contents of io stream
#         result.truncate(0)
        
        
#         ## Get file location
#         try:
#             file_loc = inspect.getfile(real_func)
#             print(file_loc)
#         except:
#             print("File location not found")
            
#         output_dict[fun_name]['file_location'] =result.getvalue()
        
        
#         ## clear contents of io stream
#         result.truncate(0)        
                
        
#     ## Reset display back to notebook
#     sys.stdout = notebook_output    

    
#     with open(json_file,'w') as f:
#         import json
#         json.dump(output_dict,f)

    
#     ## CREATE INTERACTIVE MENU
#     from ipywidgets import interact, interactive, interactive_output
#     import ipywidgets as widgets
#     from IPython.display import display
#     # from functions_combined_BEST import ihelp
#     # import functions_combined_BEST as jis

#     ## Check boxes
#     check_help = widgets.Checkbox(description="Show 'help(func)'",value=True)
#     check_source = widgets.Checkbox(description="Show source code",value=True)
#     check_fileloc=widgets.Checkbox(description="Show source filepath",value=False)
#     check_boxes = widgets.HBox(children=[check_help,check_source,check_fileloc])

#     ## dropdown menu (dropdown, label, button)
#     dropdown = widgets.Dropdown(options=list(output_dict.keys()))
#     label = widgets.Label('Function Menu')
#     button = widgets.ToggleButton(description='Show/hide',value=False)
    
#     ## Putting it all together
#     title = widgets.Label('iHelp Menu: View Help and/or Source Code')
#     menu = widgets.HBox(children=[label,dropdown,button])
#     titled_menu = widgets.VBox(children=[title,menu])
#     full_layout = widgets.GridBox(children=[titled_menu,check_boxes],box_style='warning')
    
    
    
#     ## Define output manager
#     # show_output = widgets.Output()

#     def dropdown_event(change): 
#         new_key = change.new
#         output_display = output_dict[new_key]
#     dropdown.observe(dropdown_event,names='values')

    
#     def show_ihelp(display_help=button.value,function=dropdown.value,
#                    show_help=check_help.value,show_code=check_source.value, show_file=check_fileloc.value):#,
#                    #ouput_dict=output_dict):

#         from IPython.display import Markdown
#         # import functions_combined_BEST as ji
#         from IPython.display import display        
#         page_header = '---'*28
#         import json
#         with open(json_file,'r') as f:
#             output_dict = json.load(f)
        
        
#         func_dict = output_dict[function]

#         if display_help:
#             if show_help:
# #                 display(print(func_dict['help']))
#                 print(page_header)
#                 banner = ''.join(["---"*2,' HELP ',"---"*24,'\n'])
#                 print(banner)
#                 print(func_dict['help'])

#             if show_code:
#                 print(page_header)

#                 banner = ''.join(["---"*2,' SOURCE -',"---"*23])
#                 print(banner)
#                 source_code = "```python\n"
#                 source_code += func_dict['source']
#                 source_code += "```"
#                 display(Markdown(source_code))
            
            
#             if show_file:
#                 print(page_header)
#                 banner = ''.join(["---"*2,' FILE LOCATION ',"---"*21])
#                 print(banner)
                
#                 file_loc = func_dict['file_location']
#                 print(file_loc)
                
#             if show_help==False & show_code==False & show_file==False:
#                 display('Check at least one "Show" checkbox for output.')
                
#         else:
#             display('Press "Show/hide" for display')
            
#     ## Fully integrated output
#     output = widgets.interactive_output(show_ihelp,{'display_help':button,
#                                                    'function':dropdown,
#                                                    'show_help':check_help,
#                                                    'show_code':check_source,
#                                                    'show_file':check_fileloc})

#     if to_embed:
#         return full_layout, output
#     else:
#         display(full_layout, output)


def ihelp_menu(function_list, json_file='ihelp_output.txt',to_embed=False):
    """Accepts a list of string names for loaded modules/functions to save the `help` output and 
    inspect.getsource() outputs to dictionary for later reference and display"""
    ## One way using sys to write txt file
    import pandas as pd
    import sys
    import inspect
    from io import StringIO
    notebook_output = sys.stdout
    result = StringIO()
    sys.stdout=result
    
    ## Turn single input into a list
    if isinstance(function_list,list)==False:
        function_list = [function_list]
    
    ## Make a dictionary of{function_name : function_object}
    functions_dict = dict()
    for fun in function_list:
        
        ## if input is a string, save string as key, and eval(function) as value
        if isinstance(fun, str):
            functions_dict[fun] = eval(fun)

        ## if input is a function, get the name of function using inspect and make key, function as value
        elif inspect.isfunction(fun):

            members= inspect.getmembers(fun)
            member_df = pd.DataFrame(members,columns=['param','values']).set_index('param')

            fun_name = member_df.loc['__name__'].values[0]
            functions_dict[fun_name] = fun
            
            
    ## Create an output dict to store results for functions
    output_dict = {}

    for fun_name, real_func in functions_dict.items():
        
        output_dict[fun_name] = {}
                
        ## First save help
        help(real_func)
        output_dict[fun_name]['help'] = result.getvalue()
        
        ## Clear contents of io stream
        result.truncate(0)
        
        try:
            ## Next save source
            print(inspect.getsource(real_func)) #eval(fun)))###f"{eval(fun)}"))
        except:
            print("Source code for object was not found")
        output_dict[fun_name]['source'] = result.getvalue()
        
        ## clear contents of io stream
        result.truncate(0)
        
        
        ## Get file location
        try:
            file_loc = inspect.getfile(real_func)
            print(file_loc)
        except:
            print("File location not found")
            
        output_dict[fun_name]['file_location'] =result.getvalue()
        
        
        ## clear contents of io stream
        result.truncate(0)        
                
        
    ## Reset display back to notebook
    sys.stdout = notebook_output    

    
    with open(json_file,'w') as f:
        import json
        json.dump(output_dict,f)

    
    ## CREATE INTERACTIVE MENU
    from ipywidgets import interact, interactive, interactive_output
    import ipywidgets as widgets
    from IPython.display import display
    # from functions_combined_BEST import ihelp
    # import functions_combined_BEST as ji

    ## Check boxes
    check_help = widgets.Checkbox(description="Show 'help(func)'",value=True)
    check_source = widgets.Checkbox(description="Show source code",value=True)
    check_fileloc=widgets.Checkbox(description="Show source filepath",value=False)
    check_boxes = widgets.HBox(children=[check_help,check_source,check_fileloc])

    ## dropdown menu (dropdown, label, button)
    dropdown = widgets.Dropdown(options=list(output_dict.keys()))
    label = widgets.Label('Function Menu')
    button = widgets.ToggleButton(description='Show/hide',value=False)
    
    ## Putting it all together
    title = widgets.Label('iHelp Menu: View Help and/or Source Code')
    menu = widgets.HBox(children=[label,dropdown,button])
    titled_menu = widgets.VBox(children=[title,menu])
    full_layout = widgets.GridBox(children=[titled_menu,check_boxes],box_style='warning')
    
    
    
    ## Define output manager
    # show_output = widgets.Output()

    def dropdown_event(change): 
        new_key = change.new
        output_display = output_dict[new_key]
    dropdown.observe(dropdown_event,names='values')

    
    def show_ihelp(display_help=button.value,function=dropdown.value,
                   show_help=check_help.value,show_code=check_source.value, show_file=check_fileloc.value):#,
                   #ouput_dict=output_dict):

        from IPython.display import Markdown
        # import functions_combined_BEST as ji
        from IPython.display import display        
        page_header = '---'*28
        import json
        with open(json_file,'r') as f:
            output_dict = json.load(f)
        
        
        func_dict = output_dict[function]

        if display_help:
            if show_help:
#                 display(print(func_dict['help']))
                print(page_header)
                banner = ''.join(["---"*2,' HELP ',"---"*24,'\n'])
                print(banner)
                print(func_dict['help'])

            if show_code:
                print(page_header)

                banner = ''.join(["---"*2,' SOURCE -',"---"*23])
                print(banner)
                source_code = "```python\n"
                source_code += func_dict['source']
                source_code += "```"
                display(Markdown(source_code))
            
            
            if show_file:
                print(page_header)
                banner = ''.join(["---"*2,' FILE LOCATION ',"---"*21])
                print(banner)
                
                file_loc = func_dict['file_location']
                print(file_loc)
                
            if show_help==False & show_code==False & show_file==False:
                display('Check at least one "Show" checkbox for output.')
                
        else:
            display('Press "Show/hide" for display')
            
    ## Fully integrated output
    output = widgets.interactive_output(show_ihelp,{'display_help':button,
                                                   'function':dropdown,
                                                   'show_help':check_help,
                                                   'show_code':check_source,
                                                   'show_file':check_fileloc})

    if to_embed:
        return full_layout, output
    else:
        display(full_layout, output)
        
        
        
def inspect_variables(local_vars = None,sort_col='size',exclude_funcs_mods=True, top_n=10,return_df=False,always_display=True,
show_how_to_delete=False,print_names=False):
    """
    Displays a dataframe of all variables and their size in memory,
    with the largest variables at the top. 
    
    Args:
        local_vars (locals(): Must call locals()  as first argument.
        sort_col (str, optional): column to sort by. Defaults to 'size'.
        top_n (int, optional): how many vars to show. Defaults to 10.
        return_df (bool, optional): If True, return df instead of just showing df.Defaults to False.
        always_display (bool, optional): Display df even if returned. Defaults to True.
        show_how_to_delete (bool, optional): Prints out code to copy-paste into cell to del vars. Defaults to False.
        print_names (bool, optional): [description]. Defaults to False.
    
    Raises:
        Exception: if locals() not passed as first arg
    
    
    Example Usage:
    # Must pass in local variables
    >> inspect_variables(locals())
    # To see command to delete list of vars"
    >> inspect_variables(locals(),show_how_to_delete=True)
    """
    
    

    import sys
    import inspect
    import pandas as pd
    from IPython.display import display
    if local_vars is None:
        raise Exception('Must pass "locals()" in function call. i.e. inspect_variables(locals())')


    glob_vars= [k for k in globals().keys()]
    loc_vars = [k for k in local_vars.keys()]

    var_list = glob_vars+loc_vars

    var_df = pd.DataFrame(columns=['variable','size','type'])

    exclude = ['In','Out']
    var_list = [x for x in var_list if (x.startswith('_') == False) and (x not in exclude)]

    i=0
    for var in var_list:#globals().items():#locals().items():

        if var in loc_vars:
            real_var = local_vars[var]
        elif var in glob_vars:
            real_var = globals()[var]
        else:
            print(f"{var} not found.")

        var_size = sys.getsizeof(real_var)

        var_type = []
        if inspect.isfunction(real_var):
            var_type = 'function'
            if exclude_funcs_mods:
                continue
        elif inspect.ismodule(real_var):
            var_type = 'module'
            if exclude_funcs_mods:
                continue
        elif inspect.isbuiltin(real_var):
            var_type = 'builtin'
        elif inspect.isclass(real_var):
            var_type = 'class'
        else:

            var_type = real_var.__class__.__name__


        var_row = pd.Series({'variable':var,'size':var_size,'type':var_type})
        var_df.loc[i] = var_row#pd.concat([var_df,var_row],axis=0)#.join(var_row,)
        i+=1

    # if exclude_funcs_mods:
    #     var_df = var_df.loc[var_df['type'] not in ['function', 'module'] ]

    var_df.sort_values(sort_col,ascending=False,inplace=True)
    var_df.reset_index(inplace=True,drop=True)
    var_df.set_index('variable',inplace=True)
    var_df = var_df[['type','size']]

    if top_n is not None:
        var_df = var_df.iloc[:top_n]



    if always_display:
        display(var_df.style.set_caption('Current Variables by Size in Memory'))

    if show_how_to_delete:
        print('---'*15)
        print('## CODE TO DELETE MANY VARS AT ONCE:')
        show_del_me_code(called_by_inspect_vars=True)


    if print_names ==False:
        print('#[i] set `print_names=True` for var names to copy/paste.')
        print('---'*15)
    else:
        print('---'*15)
        print('Variable Names:\n')
        print_me = [f"{str(x)}" for x in var_df.index]
        print(print_me)
    
        
    if show_del_me_code == False:
        print("[i] set `show_del_me_code=True prints copy/paste var deletion code.")
        

    if return_df:
        return var_df




def show_del_me_code(called_by_inspect_vars=False):
    """Prints code to copy and paste into a cell to delete vars using a list of their names.
    Companion function inspect_variables(locals(),print_names=True) will provide var names tocopy/paste """
    from pprint import pprint
    if called_by_inspect_vars==False:
        print("#[i]Call: `inspect_variables(locals(), print_names=True)` for list of var names")

    del_me = """
    del_me= []#list of variable names
    for me in del_me:
        try:
            exec(f'del {me}')
            print(f'del {me} succeeded')
        except:
            print(f'del {me} failed')
            continue
        """
    print(del_me)


def reload(mod):
    """Reloads the module from file without restarting kernel.
        Args:
            mod (loaded mod or list of mod objects): name or handle of package (i.e.,[ pd, fs,np])
        Returns:
            reload each model.
    Example:
    # You pass in whatever name you imported as.
    import my_functions_from_file as mf
    # after editing the source file:
    # mf.reload(mf)"""
    from importlib import reload
    print(f'Reloading...\n')
    
    ## 
    if isinstance(mod,list):
        return [reload(m) for m in mod]
    else:
        return  reload(mod)
    
    
def save_ihelp_to_file(function,save_help=False,save_code=True, 
                        as_md=False,as_txt=True,
                        folder='readme_resources/ihelp_outputs/',
                        filename=None,file_mode='w'):
    """Saves the string representation of the ihelp source code as markdown. 
    Filename should NOT have an extension. .txt or .md will be added based on
    as_md/as_txt.
    If filename is None, function name is used."""

    if as_md & as_txt:
        raise Exception('Only one of as_md / as_txt may be true.')

    import sys
    from io import StringIO
    ## save original output to restore
    orig_output = sys.stdout
    ## instantiate io stream to capture output
    io_out = StringIO()
    ## Redirect output to output stream
    sys.stdout = io_out
    
    if save_code:
        print('### SOURCE:')
        help_md = get_source_code_markdown(function)
        ## print output to io_stream
        print(help_md)
        
    if save_help:
        print('### HELP:')
        help(function)
        
    ## Get printed text from io stream
    text_to_save = io_out.getvalue()
    

    ## MAKE FULL FILENAME
    if filename is None:

        ## Find the name of the function
        import re
        func_names_exp = re.compile('def (\w*)\(')
        func_name = func_names_exp.findall(text_to_save)[0]    
        print(f'Found code for {func_name}')

        save_filename = folder+func_name#+'.txt'
    else:
        save_filename = folder+filename

    if as_md:
        ext = '.md'
    elif as_txt:
        ext='.txt'

    full_filename = save_filename + ext
    
    with open(full_filename,file_mode) as f:
        f.write(text_to_save)
        
    print(f'Output saved as {full_filename}')
    
    sys.stdout = orig_output



def get_source_code_markdown(function):
    """Retrieves the source code as a string and appends the markdown
    python syntax notation"""
    import inspect
    from IPython.display import display, Markdown
    source_DF = inspect.getsource(function)            
    output = "```python" +'\n'+source_DF+'\n'+"```"
    return output


def quick_ref(topic='student_resource_folder'):
    """Displays quick reference url links and info.
    Args:
        topic (str): selects which reference info to show.
            - `student_resource_folder` : data science gdrive url
            - `fsds` :documentaion url"""
            
    if 'student_resource_folder' in topic:
        print('Data Science Student Resources:')
        print('https://flatiron.online/StudentResourcesGdrive')
        
    if 'fsds' in topic:
        print('fsds_100719 Package Documentation:')